from datetime import datetime
from typing import Callable, Dict, Final, Generator, Iterable, Literal, Optional, Text, Tuple, Union

from ..._constants import MAX_BATCH_SIZE
from ...utils.types import B24BatchMethodTuple, JSONDict, JSONList, Timeout
from ..protocols import BitrixTokenProtocol
from ._base_caller import BaseCaller
from .call_batch import call_batch
from .call_method import call_method


class _ListFastCaller(BaseCaller):
    """"""

    _DEFAULT_ID_FIELD: Final[Text] = "ID"
    _DEFAULT_ORDER_PATTERN: Final[Callable[[Text, Text], JSONDict]] = staticmethod(lambda id_field, sorting: {"order": {id_field: sorting}})
    _HALT: Final[bool] = True
    _MAX_BATCH_SIZE: Final[int] = MAX_BATCH_SIZE
    _START: Final[int] = -1

    _REQUEST_ID_FIELDS: Final[Dict] = {
        "socialnetwork.api.workgroup": "ID",
        "tasks.task": "ID",
    }

    _ORDER_PATTERNS: Final[Dict] = {
        "department": staticmethod(lambda id_field, sorting: {"SORT": id_field, "ORDER": sorting}),
        "user": staticmethod(lambda id_field, sorting: {"SORT": id_field, "ORDER": sorting}),
        "user.userfield": _DEFAULT_ORDER_PATTERN,
    }

    __slots__ = (
        "_counter",
        "_descending",
        "_last_id",
        "_limit",
        "_now_datetime",
        "_order_pattern",
        "_request_id_field",
        "_response_id_field",
        "_results",
        "_time",
        "_wrapper",
    )

    _descending: bool
    _limit: Optional[int]
    _now_datetime: datetime
    _time: JSONDict
    _counter: int
    _last_id: int
    _request_id_field: Optional[Text]
    _response_id_field: Optional[Text]
    _wrapper: Optional[Text]
    _order_pattern: Callable[[Text, Text], JSONDict]
    _results: Optional[Union[JSONDict, JSONList]]

    def __init__(
            self,
            *,
            domain: Text,
            auth_token: Text,
            is_webhook: bool,
            api_method: Text,
            params: Optional[JSONDict] = None,
            descending: bool = False,
            limit: Optional[int] = None,
            timeout: Timeout = None,
            bitrix_token: Optional["BitrixTokenProtocol"] = None,
            **kwargs,
    ):
        super().__init__(
            domain=domain,
            auth_token=auth_token,
            is_webhook=is_webhook,
            api_method=api_method,
            params=params,
            timeout=timeout,
            bitrix_token=bitrix_token,
            **kwargs,
        )
        self._descending = descending
        self._limit = limit
        self._now_datetime = datetime.now().astimezone()
        self._time = dict(
            start=self._timesampt,
            finish=self._timesampt,
            duration=0,
            processing=0,
            date_start=self._now_datetime.isoformat(timespec="seconds"),
            date_finish=self._now_datetime.isoformat(timespec="seconds"),
        )
        self._counter = 0
        self._last_id = 0
        self._request_id_field = self._get_initial_request_id_field()
        self._response_id_field = None
        self._wrapper = None
        self._order_pattern = self._get_order_pattern()
        self._results = None

    def _get_initial_request_id_field(self) -> Optional[Text]:
        """"""

        api_method = self._api_method
        request_id_field = self._REQUEST_ID_FIELDS.get(api_method)

        while not (api_method.find(".") == -1 or request_id_field):
            api_method, _ = api_method.rsplit(".", maxsplit=1)
            request_id_field = self._REQUEST_ID_FIELDS.get(api_method)

        return request_id_field

    def _get_order_pattern(self) -> Callable[[Text, Text], JSONDict]:
        """"""

        api_method = self._api_method
        order_pattern = self._ORDER_PATTERNS.get(api_method)

        while not (api_method.find(".") == -1 or order_pattern):
            api_method, _ = api_method.rsplit(".", maxsplit=1)
            order_pattern = self._ORDER_PATTERNS.get(api_method)

        return order_pattern or self._DEFAULT_ORDER_PATTERN

    @property
    def _cmp(self) -> Literal[">", "<"]:
        """"""
        return (">", "<")[self._descending]

    @property
    def _dynamic_request_id_field(self) -> Text:
        """"""
        return self._request_id_field or self._response_id_field or self._DEFAULT_ID_FIELD

    @property
    def _prop(self) -> Text:
        """"""
        return f"{self._cmp}{self._dynamic_request_id_field}"

    @property
    def _sorting(self) -> Literal["ASC", "DESC"]:
        """"""
        return ("ASC", "DESC")[self._descending]

    @property
    def _order_by_id(self) -> JSONDict:
        """"""
        return self._order_pattern(self._dynamic_request_id_field, self._sorting)

    @property
    def _timesampt(self) -> float:
        """"""
        return self._now_datetime.timestamp()

    @staticmethod
    def _force_values(collection: Union[JSONDict, JSONList]) -> Iterable[Union[JSONDict, JSONList]]:
        """"""
        if isinstance(collection, dict):
            return collection.values()
        else:
            return collection

    @property
    def _results_values(self) -> Iterable[Union[JSONDict, JSONList]]:
        """"""
        return self._force_values(self._results)

    def _deep_merge(self, *dicts: Dict) -> Dict:
        """
        Merges nested dictionaries recursively
        """

        result_dict: Dict = dict()

        for current_dict in dicts:
            for key, value in current_dict.items():
                existing_value = result_dict.get(key)

                if isinstance(value, dict):
                    if existing_value is not None and not isinstance(existing_value, dict):
                        raise ValueError(f"Cannot merge a dict into a non-dict at key '{key}': {existing_value}")

                    result_dict[key] = self._deep_merge(existing_value or {}, value)
                else:
                    result_dict[key] = value

        return result_dict

    def _add_time(self, time: JSONDict):
        """"""

        self._time["finish"] = time["finish"]
        self._time["duration"] += time["duration"]
        self._time["processing"] += time["processing"]
        self._time["date_finish"] = time["date_finish"]

        if time.get("operating_reset_at") is not None:
            self._time["operating_reset_at"] = time["operating_reset_at"]

        if time.get("operating") is not None:
            self._time["operating"] = self._time.get("operating", 0) + time["operating"]

    def _unwrap_result(self, result: JSONDict) -> Tuple[Optional[Text], JSONList]:
        """"""

        wrapper = None

        while isinstance(result, dict):
            wrapper, result = next(iter(result.items()))

        if isinstance(result, list):
            return wrapper, result
        else:
            raise TypeError(f"API method '{self._api_method}' is not a list method!")

    def _get_path(self, index: int) -> Text:
        """"""

        path = f"$result[req_{index - 1}]"

        if self._wrapper:
            path = f"{path}[{self._wrapper}]"

        return path

    def _get_filter_by_id(self, index: int) -> JSONDict:
        """
        Generate filter by id
        """

        if index == 0:
            if self._last_id:
                return {"filter": {self._prop: self._last_id}}
            else:
                return {}

        return {"filter": {self._prop: f"{self._get_path(index)}[{self._MAX_BATCH_SIZE - 1}][{self._response_id_field}]"}}

    def _generate_method_params(self, index: int = 0) -> JSONDict:
        """"""
        return self._deep_merge(
            self._params,
            self._order_by_id,
            self._get_filter_by_id(index=index),
            dict(start=self._START),
        )

    def _generate_batch_methods(self) -> Dict[Text, B24BatchMethodTuple]:
        """
        Generates list of methods, using call_list_fast() api_method and params, adding filter by ID

        Returns:
            dict of B24BatchMethodTuple, ready to be used by call_batches()
        """

        methods: Dict[Text, B24BatchMethodTuple] = dict()

        for index in range(self._MAX_BATCH_SIZE):
            method_params = self._generate_method_params(index=index)
            methods[f"req_{index}"] = (self._api_method, method_params)

        return methods

    def _fetch_first_response(self) -> JSONDict:
        """"""
        if self._bitrix_token:
            return self._bitrix_token.call_method(
                api_method=self._api_method,
                params=self._generate_method_params(),
                timeout=self._timeout,
                **self._kwargs,
            )
        else:
            return call_method(
                domain=self._domain,
                auth_token=self._auth_token,
                is_webhook=self._is_webhook,
                api_method=self._api_method,
                params=self._generate_method_params(),
                timeout=self._timeout,
                **self._kwargs,
            )

    def _fetch_next_batch_response(self) -> JSONDict:
        """"""
        return call_batch(
                domain=self._domain,
                auth_token=self._auth_token,
                is_webhook=self._is_webhook,
                methods=self._generate_batch_methods(),
                halt=self._HALT,
                timeout=self._timeout,
                **self._kwargs,
        )

    def _extract_response_id_field(self, result_value: JSONDict) -> Text:
        """"""

        for key in result_value:
            if key.upper() == self._DEFAULT_ID_FIELD:
                return key

        raise ValueError("ID key is not found in Bitrix responses!")

    def _update_last_id(self, new_last_id: int):
        """"""
        if new_last_id != self._last_id:
            self._last_id = new_last_id
        else:
            raise ValueError(
                "Bitrix API returned the same ID sequence: "
                f"last_id={self._last_id}, new_last_id={new_last_id}. "
                "This can lead to an infinite generation loop!",
            )

    def _generate_result(self) -> Generator[JSONDict, None, None]:
        """"""

        response = self._fetch_first_response()

        self._add_time(response["time"])
        self._wrapper, unwrapped_result_values = self._unwrap_result(response["result"])

        if unwrapped_result_values:
            self._results = [response["result"]]
        else:
            return

        self._response_id_field = self._extract_response_id_field(unwrapped_result_values[0])

        while self._results:
            for result_values in self._results_values:
                unwrapped_result_values = result_values[self._wrapper] if self._wrapper else result_values

                for result_value in unwrapped_result_values:
                    yield result_value
                    self._counter += 1

                    if self._limit is not None and self._counter >= self._limit:
                        return

                if len(unwrapped_result_values) < self._MAX_BATCH_SIZE:
                    return

            new_last_id = int(unwrapped_result_values[-1][self._response_id_field])
            self._update_last_id(new_last_id)

            batch_response = self._fetch_next_batch_response()

            self._add_time(batch_response["time"])
            self._results = batch_response["result"]["result"]

    def call(self) -> JSONDict:
        """"""
        return dict(result=self._generate_result(), time=self._time)


def call_list_fast(
        *,
        domain: Text,
        auth_token: Text,
        is_webhook: bool,
        api_method: Text,
        params: Optional[JSONDict] = None,
        descending: bool = False,
        limit: Optional[int] = None,
        timeout: Timeout = None,
        bitrix_token: Optional["BitrixTokenProtocol"] = None,
        **kwargs,
) -> JSONDict:
    """
    Retrieve large number of items in a performant way using filter by ID and start=-1 parameter to disable the count of items

    Note:
        On small sets of items (2550 entries and less), call_list() has better performance and should be used instead

    Args:
        domain: bitrix portal domain
        auth_token: auth token
        is_webhook: whether the method is being called using webhook token
        api_method: name of the bitrix API method to call, e.g. crm.deal.list
        params: API method parameters
        limit: max number of items to retrieve
        descending: whether items should be retrieved in descending order
        timeout: timeout in seconds
        bitrix_token:

    Returns:
        dictionary containing list of items returned by called API method and information about call time
    """
    return _ListFastCaller(
        domain=domain,
        auth_token=auth_token,
        is_webhook=is_webhook,
        api_method=api_method,
        params=params,
        descending=descending,
        limit=limit,
        timeout=timeout,
        bitrix_token=bitrix_token,
        **kwargs,
    ).call()



from ..utils import enum as _enum


class EntityTypeAbbr(_enum.StrEnum):
    """Enumeration of CRM entity type abbreviations used in Bitrix24 CRM system."""
    LEAD = "L"
    DEAL = "D"
    CONTRACT = "C"
    COMPANY = "CO"
    INVOICE = "I"
    QUOTE = "Q"
    REQUISITE = "RQ"
    ORDER = "O"


class EntityTypeId(_enum.IntEnum):
    """Enumeration of CRM entity type IDs corresponding to Bitrix24 entities."""
    LEAD = 1
    DEAL = 2
    CONTRACT = 3
    COMPANY = 4
    INVOICE = 5
    QUOTE = 7
    REQUISITE = 8
    ORDER = 14


class EntityTypeName(_enum.StrEnum):
    """Enumeration of CRM entity type names corresponding to Bitrix24 entities."""
    LEAD = "LEAD"
    DEAL = "DEAL"
    CONTRACT = "CONTRACT"
    COMPANY = "COMPANY"
    INVOICE = "INVOICE"
    QUOTE = "QUOTE"
    REQUISITE = "REQUISITE"
    ORDER = "ORDER"


class UserFieldEntityId(_enum.StrEnum):
    """Enumeration of user field entity IDs for CRM entities in Bitrix24, used for custom fields identification."""
    LEAD = "CRM_LEAD"
    DEAL = "CRM_DEAL"
    CONTRACT = "CRM_CONTRACT"
    COMPANY = "CRM_COMPANY"
    INVOICE = "CRM_INVOICE"
    QUOTE = "CRM_QUOTE"
    REQUISITE = "CRM_REQUISITE"
    ORDER = "ORDER"



from typing import Iterable, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Recurring",
]


class Recurring(BaseCRM):
    """Methods provide capabilities for managing recurring deals.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/recurring-deals/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of fields for the recurring deal template.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/recurring-deals/crm-deal-recurring-fields.html

        The method returns a list of fields for configuring the recurring deal template with descriptions.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new recurring deal

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/recurring-deals/crm-deal-recurring-add.html

        Args:
            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                }, where each key-value pair represents a field of the deal and its corresponding value;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the fields of the recurring deal template settings by ID

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/recurring-deals/crm-deal-recurring-get.html

        The method returns the fields of the recurring deal template settings by identifier.

        Args:
            bitrix_id: Identifier of the recurring deal template settings;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Optional[int] = None,
    ) -> BitrixAPIRequest:
        """Get a list of recurring deal template settings

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/recurring-deals/crm-deal-recurring-list.html

        The method returns a list of recurring deal template settings based on the filter.

        Args:
            select: List of fields that should be populated in the selected elements;

            filter: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'ASC' (ascending sort) or 'DESC' (descending sort);
            start: This parameter is used to manage pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update existing settings for the recurring deal template

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/recurring-deals/crm-deal-recurring-update.html

        The method updates the existing settings for the recurring deal template.

        Args:
            bitrix_id: Identifier of the recurring deal template settings;

            fields: Object format:

                {
                    field_to update: value_1,

                    ...,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(bitrix_id, fields, timeout=timeout)

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete existing configuration for recurring deal template

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/recurring-deals/crm-deal-recurring-delete.html

        The method removes the existing configuration for the recurring deal template.

        Args:
            bitrix_id: Identifier of the recurring deal template configuration;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id, timeout=timeout)

    @type_checker
    def expose(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new deal from the template

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/recurring-deals/crm-deal-recurring-expose.html

        The method creates a new deal from the template.

        Args:
            bitrix_id: Identifier of the recurring deal template settings;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.expose,
            params=params,
            timeout=timeout,
        )



from functools import cached_property

from .._base_scope import BaseScope
from .activity import Activity
from .address import Address
from .automatedsolution import Automatedsolution
from .automation import Automation
from .calllist import Calllist
from .category import Category
from .company import Company
from .contact import Contact
from .currency import Currency
from .deal import Deal
from .documentgenerator import Documentgenerator
from .duplicate import Duplicate
from .entity import Entity
from .enum import Enum
from .item import Item
from .lead import Lead
from .multifield import Multifield
from .orderentity import Orderentity
from .quote import Quote
from .requisite import Requisite
from .settings import Settings
from .stagehistory import Stagehistory
from .status import Status
from .timeline import Timeline
from .type import Type
from .userfield import Userfield
from .vat import Vat

__all__ = [
    "CRM",
]


class CRM(BaseScope):
    """"""

    @cached_property
    def activity(self) -> Activity:
        """"""
        return Activity(self)

    @cached_property
    def address(self) -> Address:
        """"""
        return Address(self)

    @cached_property
    def automatedsolution(self) -> Automatedsolution:
        """"""
        return Automatedsolution(self)

    @cached_property
    def automation(self) -> Automation:
        """"""
        return Automation(self)

    @cached_property
    def calllist(self) -> Calllist:
        """"""
        return Calllist(self)

    @cached_property
    def category(self) -> Category:
        """"""
        return Category(self)

    @cached_property
    def company(self) -> Company:
        """"""
        return Company(self)

    @cached_property
    def contact(self) -> Contact:
        """"""
        return Contact(self)

    @cached_property
    def currency(self) -> Currency:
        """"""
        return Currency(self)

    @cached_property
    def deal(self) -> Deal:
        """"""
        return Deal(self)

    @cached_property
    def documentgenerator(self) -> Documentgenerator:
        """"""
        return Documentgenerator(self)

    @cached_property
    def duplicate(self) -> Duplicate:
        """"""
        return Duplicate(self)

    @cached_property
    def entity(self) -> Entity:
        """"""
        return Entity(self)

    @cached_property
    def enum(self) -> Enum:
        """"""
        return Enum(self)

    @cached_property
    def item(self) -> Item:
        """"""
        return Item(self)

    @cached_property
    def lead(self) -> Lead:
        """"""
        return Lead(self)

    @cached_property
    def multifield(self) -> Multifield:
        """"""
        return Multifield(self)

    @cached_property
    def orderentity(self) -> Orderentity:
        """"""
        return Orderentity(self)

    @cached_property
    def quote(self) -> Quote:
        """"""
        return Quote(self)

    @cached_property
    def requisite(self) -> Requisite:
        """"""
        return Requisite(self)

    @cached_property
    def settings(self) -> Settings:
        """"""
        return Settings(self)

    @cached_property
    def stagehistory(self) -> Stagehistory:
        """"""
        return Stagehistory(self)

    @cached_property
    def status(self) -> Status:
        """"""
        return Status(self)

    @cached_property
    def timeline(self) -> Timeline:
        """"""
        return Timeline(self)

    @cached_property
    def type(self) -> Type:
        """"""
        return Type(self)

    @cached_property
    def userfield(self) -> Userfield:
        """"""
        return Userfield(self)

    @cached_property
    def vat(self) -> Vat:
        """"""
        return Vat(self)



from functools import cached_property

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .._base_crm import BaseCRM
from .enumeration import Enumeration
from .settings import Settings

__all__ = [
    "Userfield",
]


class Userfield(BaseCRM):
    """Methods for working with custom fields.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/user-defined-fields/index.html
    """

    @cached_property
    def enumeration(self) -> Enumeration:
        """"""
        return Enumeration(self)

    @cached_property
    def settings(self) -> Settings:
        """"""
        return Settings(self)

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get description for custom fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/user-defined-fields/crm-userfield-fields.html

        The method returns the description of fields for custom fields.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def types(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of user field types.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/user-defined-fields/crm-userfield-types.html

        The method returns the description of fields for user fields.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.types,
            timeout=timeout,
        )



from typing import Iterable, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Localizations",
]


class Localizations(BaseCRM):
    """The methods offer capabilities for working with currency localization.

    Documenation: https://apidocs.bitrix24.com/api-reference/crm/currency/localizations/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get currency localization fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/localizations/crm-currency-localizations-fields.html

        The method retrieves the available fields for currency localization, which are settings dependent on the language.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: Text,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get currency localization.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/localizations/crm-currency-localizations-get.html

        This method retrieves existing currency localization.

        Args:
            bitrix_id: Currency symbolic identifier;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def set(
            self,
            bitrix_id: Text,
            *,
            localizations: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Set localization for currency.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/localizations/crm-currency-localizations-set.html

        This method updates localizations for a currency or adds them if the localization for the specified language does not exist.

        Args:
            bitrix_id: Currency identifier;

            localizations: Currency localization parameters;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "localizations": localizations,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.set,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: Text,
            *,
            lids: Iterable[Text],
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete currency localization.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/currency/localizations/crm-currency-localizations-delete.html

        This method removes currency localizations for the specified languages.

        Args:
            bitrix_id: Currency identifier;

            lids: Array of language identifiers for which localizations need to be removed;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        if lids.__class__ is not list:
            lids = list(lids)

        params = {
            "id": bitrix_id,
            "lids": lids,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from typing import Iterable, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Link",
]


class Link(BaseCRM):
    """The methods provide capabilities for managing links between requisites and CRM objects.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/links/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get description of CRM requisite.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/links/crm-requisite-link-fields.html

        The method returns a formal description of the link fields for requisites.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def get(
            self,
            *,
            entity_type_id: int,
            entity_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get requisite link with CRM object.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/links/crm-requisite-link-get.html

        The method returns the link between requisites and an object.

        Args:
            entity_type_id: Identifier of the object type to which the link belongs, where possible values are:

                - deal (value 2),

                - old invoice (value 5),

                - estimate (value 7),

                - new invoice (value 31),

                - other dynamic objects;

            entity_id: Identifier of the object to which the link belongs;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of requisite links.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/links/crm-requisite-link-list.html

        The method returns a list of requisite links based on the filter.

        Args:
            select: An array of fields to select;

            filter: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            start: This parameter is used for pagination control;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def register(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Register requisite link.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/links/crm-requisite-link-register.html

        This method registers a link between requisites and an object.

        Args:
            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.register,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def unregister(
            self,
            *,
            entity_type_id: int,
            entity_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Unlink requisite from object.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/links/crm-requisite-link-unregister.html

        This method removes the link between requisites and an object.

        Args:
            entity_type_id: Identifier of the object type to which the link belongs, where possible values are:

                - deal (value 2),

                - old invoice (value 5),

                - estimate (value 7),

                - new invoice (value 31),

                - other dynamic objects;

            entity_id: Identifier of the object to which the link belongs;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.unregister,
            params=params,
            timeout=timeout,
        )



import enum as _enum
import typing

from .._constants import PYTHON_VERSION as _PV

__all__ = [
    "IntEnum",
    "StrEnum",
]

if _PV >= (3, 11):
    IntEnum = _enum.IntEnum
else:
    class IntEnum(_enum.IntEnum):
        """"""

        def __str__(self) -> typing.Text:
            return str(self.value)


if _PV >= (3, 11):
    StrEnum = _enum.StrEnum
else:
    class StrEnum(str, _enum.Enum):
        """"""

        def __repr__(self) -> typing.Text:
            return f"<{self.__class__.__name__}.{self.name}: {self.value!r}>"

        def __str__(self) -> typing.Text:
            return str(self.value)



from .credentials import (
    AbstractBitrixApp,
    AbstractBitrixAppLocal,
    AbstractBitrixToken,
    AbstractBitrixTokenLocal,
    BitrixApp,
    BitrixAppLocal,
    BitrixToken,
    BitrixTokenLocal,
    BitrixWebhook,
)

__all__ = [
    "AbstractBitrixApp",
    "AbstractBitrixAppLocal",
    "AbstractBitrixToken",
    "AbstractBitrixTokenLocal",
    "BitrixApp",
    "BitrixAppLocal",
    "BitrixToken",
    "BitrixTokenLocal",
    "BitrixWebhook",
]



import logging
from typing import Iterable, Optional, Text

from .base_logger import BaseLogger


class StreamLogger(BaseLogger):
    """"""

    _DEFAULT_FMT: Text = "%(asctime)s [%(levelname)s] %(message)s"
    _DEFAULT_HANDLER_TYPE = logging.StreamHandler
    _DEFAULT_LEVEL = logging.DEBUG

    __slots__ = ("_formatter",)

    _formatter: Optional[logging.Formatter]

    class ContextFormatter(logging.Formatter):

        def format(self, record: logging.LogRecord) -> Text:
            message = super().format(record)
            context = getattr(record, "context", None)

            if context:
                message = f"{message} | context: {context}"

            return message

    def __init__(
            self,
            *,
            name: Optional[Text] = None,
            level: Optional[int] = None,
            handlers: Optional[Iterable[logging.Handler]] = None,
            fmt: Optional[Text] = None,
            formatter: Optional[logging.Formatter] = None,
    ):
        self._formatter = formatter or self.ContextFormatter(fmt or self.get_default_fmt())
        super().__init__(
            name=name,
            level=level,
            handlers=handlers,
        )

    @classmethod
    def get_default_fmt(cls) -> Text:
        return cls._DEFAULT_FMT

    @classmethod
    def get_default_formatter(cls) -> logging.Formatter:
        return cls.ContextFormatter(cls.get_default_fmt())

    @property
    def formatter(self) -> logging.Formatter:
        return self._formatter

    def set_handler(self, handler: logging.Handler):
        handler.setFormatter(self._formatter)
        super().set_handler(handler)

    def set_formatter(self, formatter: logging.Formatter):
        self._formatter = formatter

        for handler in self.handlers:
            handler.setFormatter(self._formatter)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, Optional, Text, Union

from ..bitrix_api.requests import BitrixAPIRequest
from ..utils.functional import Classproperty
from ..utils.types import JSONDict, Timeout

if TYPE_CHECKING:
    from .. import Client
    from ..bitrix_api.credentials import AbstractBitrixToken


class BaseContext(ABC):
    """"""

    __slots__ = ()

    def __str__(self):
        return self._path

    @Classproperty
    def _name(cls) -> Text:
        return cls.__name__.lower()

    @property
    @abstractmethod
    def _context(self) -> Union["BaseContext", "Client"]:
        """"""
        raise NotImplementedError

    @property
    def _bitrix_token(self) -> "AbstractBitrixToken":
        """"""
        return getattr(self._context, "_bitrix_token")

    @property
    def _kwargs(self) -> JSONDict:
        """"""
        return getattr(self._context, "_kwargs")

    @property
    def _path(self) -> Text:
        """"""
        base_path = getattr(self._context, "_path", None)
        return f"{base_path}.{self._name}" if base_path else self._name

    @staticmethod
    def __to_camel_case(snake_str: Text) -> Text:
        """Converts Python methods names to camelCase to be used in _get_api_method"""
        first, *parts = snake_str.split("_")
        return "".join([first.lower(), *(part.title() for part in parts)])

    def _get_api_method(self, api_wrapper: Callable) -> Text:
        """"""
        api_wrapper_name = getattr(api_wrapper, "__name__", None)
        return f"{self}.{self.__to_camel_case(api_wrapper_name.strip('_'))}" if api_wrapper_name else str(self)

    def _make_bitrix_api_request(
            self,
            api_wrapper: Callable,
            params: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return BitrixAPIRequest(
            bitrix_token=self._bitrix_token,
            api_method=self._get_api_method(api_wrapper),
            params=params,
            timeout=timeout,
            **self._kwargs,
        )



from typing import Optional

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Delivery",
]


class Delivery(BaseCRM):
    """These methods offer capabilities for obtaining information about deliveries.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/delivery/index.html
    """

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get delivery information by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/delivery/crm-item-delivery-get.html

        This method retrieves brief information about delivery.

        Args:
            bitrix_id: Delivery identifier;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            entity_type_id: int,
            entity_id: int,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get list of deliveries.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/delivery/crm-item-delivery-list.html

        Args:
            entity_type_id: Identifier of the crm object type;

            entity_id: Identifier of the crm object;

            filter: Object format:

                {
                    "field_1": "value_1",

                    "field_2": "value_2",

                    ...,

                    "field_n": "value_n",
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
        }

        if filter is not None:
            params["filter"] = filter

        if order is not None:
            params["order"] = order

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )



from typing import Optional

from .....bitrix_api.requests import BitrixAPIRequest
from .....utils.functional import type_checker
from .....utils.types import JSONDict, Timeout
from ..._base_crm import BaseCRM

__all__ = [
    "Product",
]


class Product(BaseCRM):
    """These methods offer capabilities for managing product items in payments.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/products-in-payment/index.html
    """

    @type_checker
    def add(
            self,
            *,
            payment_id: int,
            row_id: int,
            quantity: float,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add product item to payment.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/products-in-payment/crm-item-payment-product-add.html

        This method adds a product item to the payment.

        Args:
            payment_id: Payment identifier;

            row_id: Identifier of the product item in the CRM object;

            quantity: Quantity of the product;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "paymentId": payment_id,
            "rowId": row_id,
            "quantity": quantity,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            payment_id: int,
            filter: JSONDict,
            order: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get list of payment product items.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/products-in-payment/crm-item-payment-product-list.html

        This method retrieves a list of product items (goods or services) associated with a specific payment.

        Args:
            payment_id: Identifier of the payment;

            filter: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "paymentId": payment_id,
            "filter": filter,
        }

        if order is not None:
            params["order"] = order

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Remove product item from payment.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/products-in-payment/crm-item-payment-product-delete.html

        This method removes a product item from the payment.

        Args:
            bitrix_id: Identifier of the product item in the payment;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id, timeout=timeout)

    @type_checker
    def set_quantity(
            self,
            bitrix_id: int,
            *,
            quantity: float,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Change product quantity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/products-in-payment/crm-item-payment-product-set-quantity.html

        This method changes the quantity of a product in the payment line item.

        Args:
            bitrix_id: Identifier of the product line item in the payment;

            quantity: Quantity of the product;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "quantity": quantity,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.set_quantity,
            params=params,
            timeout=timeout,
        )



from typing import IO, Dict, Optional, Text, Tuple

from ...utils.types import JSONDict, Timeout
from ..requesters import BitrixAPIRequester


def call(
        url: Text,
        *,
        params: Optional[JSONDict] = None,
        files: Optional[Dict[Text, Tuple[Text, IO]]] = None,
        timeout: Timeout = None,
        max_retries: Optional[int] = None,
        initial_retry_delay: Optional[float] = None,
        retry_delay_increment: Optional[float] = None,
) -> JSONDict:
    """
    Performs a call to the Bitrix API

    Args:
        url: url to which the requests should be sent
        params: API method parameters
        files: files attached to the requests
        timeout: timeout in seconds
        max_retries: maximum number of retries that will occur when server is not responding
        initial_retry_delay: initial delay between retries in seconds
        retry_delay_increment: amount by which delay between retries will increment after each retry

    Returns:
        Response returned by the API server
    Raises:
        BitrixRequestError: if failed to establish HTTP connection
        BitrixRequestTimeout: if the requests timed out
    """
    return BitrixAPIRequester(
        url=url,
        params=params,
        files=files,
        timeout=timeout,
        max_retries=max_retries,
        initial_retry_delay=initial_retry_delay,
        retry_delay_increment=retry_delay_increment,
    ).call()



from functools import cached_property

from ....bitrix_api.requests import BitrixAPIRequest
from ....scopes.crm._base_crm import BaseCRM
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .settings import Settings

__all__ = [
    "Enum",
]


class Enum(BaseCRM):
    """The methods return information about the values of types: address type, activity type, object type, and others.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/enum/index.html
    """

    @cached_property
    def settings(self) -> Settings:
        """"""
        return Settings(self)

    @type_checker
    def activitydirection(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get activity direction enumeration elements.

        DocumentationL https://apidocs.bitrix24.com/api-reference/crm/auxiliary/enum/outdated/crm-enum-activity-direction.html

        The method returns activity directions for the DIRECTION field of deals, emails, and calls.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.activitydirection,
            timeout=timeout,
        )

    @type_checker
    def activitynotifytype(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get enumeration items 'Activity notification type'.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/enum/outdated/crm-enum-activity-notify-type.html

        The method returns notification types for the NOTIFY_TYPE field of meetings and calls.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.activitynotifytype,
            timeout=timeout,
        )

    @type_checker
    def activitypriority(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get enumeration items 'Activity priority'.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/enum/outdated/crm-enum-activity-priority.html

        The method returns a list of properties for the PRIORITY field of deals.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.activitypriority,
            timeout=timeout,
        )

    @type_checker
    def activitystatus(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get enumeration items 'Status'.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/enum/outdated/crm-enum-activity-status.html

        The method returns a list of statuses for the STATUS field of deals.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.activitystatus,
            timeout=timeout,
        )

    @type_checker
    def activitytype(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get enumeration item 'Activity types'.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/enum/outdated/crm-enum-activity-type.html

        The method returns a list of types for the TYPE_ID field of activities.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.activitytype,
            timeout=timeout,
        )

    @type_checker
    def addresstype(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get address types.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/enum/crm-enum-address-type.html

        The method returns a list of address types.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.addresstype,
            timeout=timeout,
        )

    @type_checker
    def contenttype(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get enumeration items for 'Description type'.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/enum/outdated/crm-enum-content-type.html

        The method returns description types for the DESCRIPTION_TYPE field of deals.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.contenttype,
            timeout=timeout,
        )

    @type_checker
    def getorderownertypes(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get object types for order binding.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/enum/crm-enum-get-order-owner-types.html

        The method returns a list of object types to which an order can be bound.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.getorderownertypes,
            timeout=timeout,
        )

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get fields of CRM enumeration elements.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/enum/crm-enum-fields.html

        The method returns information about the fields of enumeration elements.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def ownertype(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get CRM object types.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/auxiliary/enum/crm-enum-owner-type.html

        The method returns the identifiers of CRM object types and smart processes.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._make_bitrix_api_request(
            api_wrapper=self.ownertype,
            timeout=timeout,
        )



from functools import cached_property
from typing import Iterable, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import B24Bool, JSONDict, Timeout
from .base_item import BaseItem
from .delivery import Delivery
from .details import Details
from .payment import Payment
from .productrow import Productrow

__all__ = [
    "Item",
]


class Item(BaseItem):
    """The methods provide capabilities for managing various CRM entities, such as leads, deals, contacts, companies, invoices, estimates, and SPA elements.
    They allow you to retrieve fields, add, update, delete, and get lists of elements.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/index.html
    """

    @cached_property
    def delivery(self) -> "Delivery":
        """"""
        return Delivery(self)

    @cached_property
    def details(self) -> "Details":
        """"""
        return Details(self)

    @cached_property
    def payment(self) -> "Payment":
        """"""
        return Payment(self)

    @cached_property
    def productrow(self) -> "Productrow":
        """"""
        return Productrow(self)

    @type_checker
    def fields(
            self,
            *,
            entity_type_id: int,
            use_original_uf_names: Optional[bool] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get fields of CRM item.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/crm-item-fields.html

        This method retrieves a list of fields and their configuration for items of type entityTypeId.

        Args:
            entity_type_id: Identifier of the system or custom type whose element we want to retrieve;

            use_original_uf_names: This parameter controls the format of custom field names in the responses;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(
            entity_type_id=entity_type_id,
            use_original_uf_names=use_original_uf_names,
            timeout=timeout,
        )

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            entity_type_id: int,
            use_original_uf_names: Optional[bool] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new CRM entity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/crm-item-add.html

        This method is a universal way to create objects in CRM. With it, you can create various types of objects, such as deals, contacts, companies, and more.

        To create an object, you need to pass the appropriate parameters, including the object type and its information: name, description, contact details, and other specifics.

        Upon successful execution of the requests, a new object is created.

        Args:
            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            entity_type_id: Identifier of the system or user-defined type whose element we want to create;

            use_original_uf_names: This parameter controls the format of custom field names in the responses;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(
            fields,
            entity_type_id=entity_type_id,
            use_original_uf_names=use_original_uf_names,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            entity_type_id: int,
            use_original_uf_names: Optional[bool] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get an item by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/crm-item-get.html

        The method returns information about an item based on the item identifier and the CRM object type identifier.

        Args:
            bitrix_id: Identifier of the item whose information we want to obtain;

            entity_type_id: Identifier of the system or user-defined type whose item we want to retrieve;

            use_original_uf_names: This parameter is used to control the format of custom field names in the responses;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(
            bitrix_id=bitrix_id,
            entity_type_id=entity_type_id,
            use_original_uf_names=use_original_uf_names,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            entity_type_id: int,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            use_original_uf_names: Optional[bool] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of CRM elements.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/crm-item-list.html

        This method retrieves a list of elements of a specific type of CRM entity.

        CRM entity elements will not be included in the final selection if the user does not have "read" access permission for these elements.

        Args:
            entity_type_id: Identifier of the system or user-defined type whose item we want to retrieve;

            select: List of fields that should be populated in the selected elements;

            filter: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'ASC' (ascending sort) or 'DESC' (descending sort);

            start: This parameter is used to manage pagination;

            use_original_uf_names: This parameter controls the format of user field names in the requests and responses;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            entity_type_id=entity_type_id,
            select=select,
            filter=filter,
            order=order,
            start=start,
            use_original_uf_names=use_original_uf_names,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            entity_type_id: int,
            use_original_uf_names: Optional[bool] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update CRM item.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/crm-item-update.html

        This method updates an item of a specific type in the CRM object by assigning new values from the fields parameter.

        Args:
            bitrix_id: Identifier of the item we want to change;

            fields: Object format

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            entity_type_id: Identifier of the system or user-defined type whose item we want to change;

            use_original_uf_names: Parameter to control the format of custom field names in the requests and responses;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id,
            fields,
            entity_type_id=entity_type_id,
            use_original_uf_names=use_original_uf_names,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            entity_type_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete CRM item.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/crm-item-delete.html

        This method deletes a CRM entity item by its item ID and entity type ID.

        Args:
            bitrix_id: The ID of the item to be deleted;

            entity_type_id: The ID of the system or user-defined type of the item we want to delete;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(
            bitrix_id=bitrix_id,
            entity_type_id=entity_type_id,
            timeout=timeout,
        )

    @type_checker
    def import_(
            self,
            fields: JSONDict,
            *,
            entity_type_id: int,
            use_original_uf_names: Optional[bool] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Import a single record

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/import/crm-item-import.html

        A universal method for importing objects into CRM.

        Args:
            fields: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,

                }

                where:
                    field_n — field name

                    value_n — field value;

            entity_type_id: Identifier of the system or custom type for which the item needs to be created;

            use_original_uf_names: Parameter to control the format of custom field names in the requests and responses;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "fields": fields,
        }

        if use_original_uf_names is not None:
            params["useOriginalUfNames"] = B24Bool(use_original_uf_names).to_str()

        return self._make_bitrix_api_request(
            api_wrapper=self.import_,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def batch_import(
            self,
            data: Iterable[JSONDict],
            *,
            entity_type_id: int,
            use_original_uf_names: Optional[bool] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Import a batch of CRM records.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/import/crm-item-batch-import.html

        A universal method for importing objects into CRM.

        Args:
            data: An array of fields values for the items;

            entity_type_id: Identifier of the system or custom type for which the item needs to be created;

            use_original_uf_names: Parameter to control the format of custom field names in the requests and responses;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        if data.__class__ is not list:
            data = list(data)

        params = {
            "entityTypeId": entity_type_id,
            "data": data,
        }

        if use_original_uf_names is not None:
            params["useOriginalUfNames"] = B24Bool(use_original_uf_names).to_str()

        return self._make_bitrix_api_request(
            api_wrapper=self.batch_import,
            params=params,
            timeout=timeout,
        )



import json
from dataclasses import asdict, dataclass
from typing import Dict, Literal, Optional, Text, cast

from ..._constants import PYTHON_VERSION as _PV
from ...constants import B24AppStatus
from ...error import BitrixValidationError
from ...utils.types import JSONDict
from .oauth_token import OAuthToken

_DATACLASS_KWARGS = {"eq": False, "frozen": True}

if _PV >= (3, 10):
    _DATACLASS_KWARGS["slots"] = True


@dataclass(**_DATACLASS_KWARGS)
class OAuthPlacementData:
    """"""

    class ValidationError(BitrixValidationError):
        """"""

    oauth_token: OAuthToken
    domain: Text
    protocol: Literal[0, 1]
    lang: Text
    app_sid: Text
    member_id: Text
    status: B24AppStatus
    placement: Optional[Text]
    placement_options: Optional[JSONDict]

    @classmethod
    def from_dict(cls, placement_data: JSONDict) -> "OAuthPlacementData":
        try:
            oauth_token = OAuthToken.from_placement_data(placement_data)

            domain = placement_data["DOMAIN"]
            protocol = int(placement_data["PROTOCOL"])
            lang = placement_data["LANG"]
            app_sid = placement_data["APP_SID"]
            member_id = placement_data["member_id"]
            status = B24AppStatus(placement_data["status"])
            placement = placement_data.get("PLACEMENT")
            placement_options = placement_data.get("PLACEMENT_OPTIONS") and json.loads(placement_data["PLACEMENT_OPTIONS"])

            return cls(
                oauth_token=oauth_token,
                domain=domain,
                protocol=cast(Literal[0, 1], protocol),
                lang=lang,
                app_sid=app_sid,
                member_id=member_id,
                status=status,
                placement=placement,
                placement_options=placement_options,
            )

        except KeyError as error:
            raise cls.ValidationError(f"Missing required field in placement data: {error.args[0]}") from error

        except Exception as error:
            raise cls.ValidationError(f"Invalid placement data: {error}") from error

    def to_dict(self) -> Dict:
        return asdict(self)



from datetime import datetime
from typing import List, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Todo",
]


class Todo(BaseCRM):
    """These methods offer capabilities for managing universal activities, which are a type of activity with extended settings.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/todo/index.html
    """

    @type_checker
    def add(
            self,
            *,
            owner_type_id: int,
            owner_id: int,
            deadline: datetime,
            title: Optional[Text] = None,
            description: Optional[Text] = None,
            responsible_id: Optional[int] = None,
            parent_activity_id: Optional[int] = None,
            ping_offsets: Optional[List[int]] = None,
            color_id: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add a new universal activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/todo/crm-activity-todo-add.html

        The method adds a universal activity to the timeline.

        Args:
            owner_type_id: Identifier of the CRM object type to which the activity is linked;

            owner_id: Identifier of the CRM entity to which the activity is linked;

            deadline: Deadline for the activity;

            title: Title of the activity, default is an empty string;

            description: Description of the activity, default is an empty string;

            responsible_id: Identifier of the user responsible for the activity;

            parent_activity_id: Identifier of the activity in the timeline with which the created activity can be linked;

            ping_offsets: An array containing integer values in minutes that allow you to set reminder times for the activity;

            color_id: Identifier of the activity color in the timeline. There are 8 colors available, values from 1 to 7 and a default color if none is specified;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "ownerTypeId": owner_type_id,
            "ownerId": owner_id,
            "deadline": deadline,
            "title": title,
            "description": description,
            "colorId": color_id,
        }

        if ping_offsets is not None:
            params["pingOffsets"] = ping_offsets

        if responsible_id is not None:
            params["responsibleId"] = responsible_id

        if parent_activity_id is not None:
            params["parentActivityId"] = parent_activity_id

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            *,
            owner_type_id: int,
            owner_id: int,
            deadline: datetime,
            title: Optional[Text] = None,
            description: Optional[Text] = None,
            responsible_id: Optional[int] = None,
            parent_activity_id: Optional[int] = None,
            ping_offsets: Optional[List[int]] = None,
            color_id: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update universal activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/todo/crm-activity-todo-update.html

        The method updates a universal activity.

        Args:
            bitrix_id: Identifier of the activity;

            owner_type_id: Identifier of the CRM object type to which the activity is linked;

            owner_id: Identifier of the CRM entity to which the activity is linked;

            deadline: Deadline for the activity;

            title: Title of the activity, default is an empty string;

            description: Description of the activity, default is an empty string;

            responsible_id: Identifier of the user responsible for the activity;

            parent_activity_id: Identifier of the activity in the timeline with which the created activity can be linked;

            ping_offsets: An array containing integer values in minutes that allow you to set reminder times for the activity;

            color_id: Identifier of the activity color in the timeline. There are 8 colors available, values from 1 to 7 and a default color if none is specified;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "ownerTypeId": owner_type_id,
            "ownerId": owner_id,
            "deadline": deadline,
            "title": title,
            "description": description,
            "colorId": color_id,
        }

        if ping_offsets is not None:
            params["pingOffsets"] = ping_offsets

        if responsible_id is not None:
            params["responsibleId"] = responsible_id

        if parent_activity_id is not None:
            params["parentActivityId"] = parent_activity_id

        return self._make_bitrix_api_request(
            api_wrapper=self.update,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def update_color(
            self,
            bitrix_id: int,
            *,
            owner_type_id: int,
            owner_id: int,
            color_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update the color of the universal activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/todo/crm-activity-todo-update-color.html

        The method updates the color of the universal activity.

        Args:
            bitrix_id: Identifier of the activity being updated;

            owner_type_id: Identifier of the CRM object type to which the activity is linked;

            owner_id: Identifier of the CRM entity to which the activity is linked;

            color_id: Identifier of the activity color in the timeline. There are 8 colors available, values from 1 to 7 and the default color if none is specified;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "ownerTypeId": owner_type_id,
            "ownerId": owner_id,
            "colorId": color_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.update_color,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def update_deadline(
            self,
            bitrix_id: int,
            *,
            owner_type_id: int,
            owner_id: int,
            value: datetime,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update the deadline of the universal activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/todo/crm-activity-todo-update-deadline.html

        The method changes the deadline of a universal activity.

        Args:
            bitrix_id: Identifier of the activity being updated;

            owner_type_id: Identifier of the CRM object type to which the activity is linked;

            owner_id: Identifier of the CRM entity to which the activity is linked;

            value: New deadline for the activity;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "ownerTypeId": owner_type_id,
            "ownerId": owner_id,
            "value": value,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.update_deadline,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def update_description(
            self,
            bitrix_id: int,
            *,
            owner_type_id: int,
            owner_id: int,
            value: Text,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update the description of the universal activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/todo/crm-activity-todo-update-description.html

        The method changes the description in the universal activity.

        Args:
            bitrix_id: Identifier of the activity being updated;

            owner_type_id: Identifier of the CRM object type to which the activity is linked;

            owner_id: Identifier of the CRM entity to which the activity is linked;

            value: New description of the activity;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "ownerTypeId": owner_type_id,
            "ownerId": owner_id,
            "value": value,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.update_description,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def update_responsible_user(
            self,
            bitrix_id: int,
            *,
            owner_type_id: int,
            owner_id: int,
            responsible_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update the responsible user for the universal activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/todo/crm-activity-todo-update-responsible-user.html

        The method updates the responsible user for the universal activity.

        Args:
            bitrix_id: Identifier of the activity being updated;

            owner_type_id: Identifier of the CRM object type to which the activity is linked;

            owner_id: Identifier of the CRM entity to which the activity is linked;

            responsible_id: Identifier of the user who will become responsible for the activity;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "ownerTypeId": owner_type_id,
            "ownerId": owner_id,
            "responsibleId": responsible_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.update_responsible_user,
            params=params,
            timeout=timeout,
        )



from functools import cached_property
from typing import Iterable, Optional, Text

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from ._relationships import Contact
from ._userfield import Userfield
from .details import Details
from .item.base_item import BaseItem

__all__ = [
    "Company",
]


class Company(BaseItem):
    """The methods provide capabilities for managing companies.
    They allow you to retrieve fields, add, update, delete, and get lists of companies.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/companies/index.html
    """

    @cached_property
    def contact(self) -> Contact:
        """"""
        return Contact(self)

    @cached_property
    def details(self) -> "Details":
        """"""
        return Details(self)

    @cached_property
    def userfield(self) -> Userfield:
        """"""
        return Userfield(self)

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get company fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/companies/crm-company-fields.html

        The method returns the description of company fields? including custom fields.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            params: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new company.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/companies/crm-company-add.html

        The method create a new company.

        Args:
            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            params: An objects containing a set of additional parameters where

                - REGISTER_SONET_EVENT - whether to register the change event in the live feed 'Y' or not 'N'

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(
            fields,
            extra_params=params,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get company by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/companies/crm-company-get.html

        The method returns a company by its identifier.

        Args:
            bitrix_id: Identifier of the company;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of companies.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/companies/crm-company-list.html

        The method returns a list of companies based on a filter.

        Args:
            select: List of fields that should be populated in the selected elements;

            filter: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'ASC' (ascending sort) or 'DESC' (descending sort);

            start: This parameter is used to manage pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            params: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update company.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/companies/crm-company-update.html

        The method updates an existing company.

        Args:
            bitrix_id: Identifier of the company to be changed;

            fields: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            params: Set of additional parameters where

                - REGISTER_SONET_EVENT - whether to register the change event in the live feed 'Y' or not 'N';

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id,
            fields,
            extra_params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete company.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/companies/crm-company-delete.html

        The method removes a contact and all associated objects.

        Args:
            bitrix_id: Identifier of the company;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id, timeout=timeout)



from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from typing import Dict, Optional, Text

from ..._constants import PYTHON_VERSION as _PV
from ...error import BitrixValidationError
from ...utils.types import JSONDict

_DATACLASS_KWARGS = {"eq": False, "frozen": True}

if _PV >= (3, 10):
    _DATACLASS_KWARGS["slots"] = True


@dataclass(**_DATACLASS_KWARGS)
class OAuthToken:
    """"""

    class ValidationError(BitrixValidationError):
        """"""

    access_token: Text
    refresh_token: Optional[Text]
    expires: Optional[datetime]
    expires_in: Optional[int] = None

    @classmethod
    def from_dict(cls, oauth_token_payload: JSONDict) -> "OAuthToken":
        """"""
        try:
            return cls(
                access_token=oauth_token_payload["access_token"],
                refresh_token=oauth_token_payload["refresh_token"],
                expires=datetime.fromtimestamp(int(oauth_token_payload["expires"])).astimezone(),
                expires_in=int(oauth_token_payload["expires_in"]),
            )
        except KeyError as error:
            raise cls.ValidationError(f"Missing required field in OAuth token payload: {error.args[0]}") from error
        except Exception as error:
            raise cls.ValidationError(f"Invalid OAuth token payload: {error}") from error

    @classmethod
    def from_placement_data(cls, placement_data: JSONDict) -> "OAuthToken":
        """"""

        try:
            access_token = placement_data["AUTH_ID"]
            refresh_token = placement_data["REFRESH_ID"]
            expires_in = int(placement_data["AUTH_EXPIRES"])
            expires = datetime.now().astimezone() + timedelta(seconds=expires_in)

            return cls(
                access_token=access_token,
                refresh_token=refresh_token,
                expires=expires,
                expires_in=expires_in,
            )

        except KeyError as error:
            raise cls.ValidationError(f"Missing required field in OAuth token placement data: {error.args[0]}") from error

        except Exception as error:
            raise cls.ValidationError(f"Invalid OAuth token placement data: {error}") from error

    @property
    def is_one_off(self) -> bool:
        """"""
        return bool(self.refresh_token)

    @property
    def has_expired(self) -> Optional[bool]:
        """"""
        return self.expires and self.expires <= datetime.now().astimezone()

    def to_dict(self) -> Dict:
        return asdict(self)



import logging
from abc import ABC, abstractmethod
from typing import Any, ClassVar, Dict, Mapping, Optional, Text


class AbstractLogger(ABC):
    """"""

    NOTSET = logging.NOTSET
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    LOG_LEVELS: ClassVar[Dict[Text, int]] = {
        "NOTSET": NOTSET,
        "DEBUG": DEBUG,
        "INFO": INFO,
        "WARNING": WARNING,
        "ERROR": ERROR,
        "CRITICAL": CRITICAL,
    }

    @property
    @abstractmethod
    def level(self) -> int: ...

    @abstractmethod
    def debug(self, message: Text, context: Optional[Mapping[Text, Any]] = None): ...

    @abstractmethod
    def info(self, message: Text, context: Optional[Mapping[Text, Any]] = None): ...

    @abstractmethod
    def warning(self, message: Text, context: Optional[Mapping[Text, Any]] = None): ...

    @abstractmethod
    def error(self, message: Text, context: Optional[Mapping[Text, Any]] = None): ...

    @abstractmethod
    def critical(self, message: Text, context: Optional[Mapping[Text, Any]] = None): ...

    @abstractmethod
    def log(self, level: int, message: Text, context: Optional[Mapping[Text, Any]] = None): ...

    @abstractmethod
    def set_level(self, level: int): ...



from functools import cached_property

from .._base_crm import BaseCRM
from .mode import Mode

__all__ = [
    "Settings",
]


class Settings(BaseCRM):
    """"""

    @cached_property
    def mode(self) -> Mode:
        """"""
        return Mode(self)



from dataclasses import asdict, dataclass, field
from typing import Dict, List, Literal, Text, Union

from ..._constants import PYTHON_VERSION as _PV
from ...utils.types import B24APIResult, JSONDict, JSONList
from .bitrix_api_response import BitrixAPIResponse
from .bitrix_api_time_response import BitrixAPITimeResponse

_DATACLASS_KWARGS = {"repr": False, "eq": False, "frozen": True}

if _PV >= (3, 10):
    _DATACLASS_KWARGS["slots"] = True


@dataclass(**_DATACLASS_KWARGS)
class B24APIBatchResult:
    """"""

    result: Union[Dict[Text, B24APIResult], List[B24APIResult]]
    result_error: Union[JSONDict, JSONList]
    result_total: Union[Dict[Text, int], List[int]]
    result_next: Union[Dict[Text, int], List[int]]
    result_time: Union[Dict[Text, BitrixAPITimeResponse], List[BitrixAPITimeResponse]]

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"result=<{type(self.result).__name__}: {len(self.result)}>, "
            f"result_error=<{type(self.result_error).__name__}: {len(self.result_error)}>, "
            f"result_total=<{type(self.result_total).__name__}: {len(self.result_total)}>, "
            f"result_next=<{type(self.result_next).__name__}: {len(self.result_next)}>, "
            f"result_time=<{type(self.result_time).__name__}: {len(self.result_time)}>)"
        )

    @classmethod
    def from_dict(cls, json_response: JSONDict) -> "B24APIBatchResult":
        json_result_time = json_response["result_time"]

        if isinstance(json_result_time, dict):
            result_time = dict()

            for key, time_value in json_result_time.items():
                result_time[key] = BitrixAPITimeResponse.from_dict(time_value)

        else:
            result_time = list()

            for time_value in json_result_time:
                result_time.append(BitrixAPITimeResponse.from_dict(time_value))

        return cls(
            result=json_response["result"],
            result_error=json_response["result_error"],
            result_total=json_response["result_total"],
            result_next=json_response["result_next"],
            result_time=result_time,
        )

    def to_dict(self) -> Dict:
        return asdict(self)


@dataclass(**_DATACLASS_KWARGS)
class BitrixAPIBatchResponse(BitrixAPIResponse):
    """"""

    result: B24APIBatchResult
    next: Literal[None] = field(init=False, default=None)
    total: Literal[None] = field(init=False, default=None)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"result={self.result}, "
            f"time={self.time})"
        )

    @classmethod
    def from_dict(cls, json_response: JSONDict) -> "BitrixAPIBatchResponse":
        return cls(
            result=B24APIBatchResult.from_dict(json_response["result"]),
            time=BitrixAPITimeResponse.from_dict(json_response["time"]),
        )



from typing import Iterable, Text

from ..bitrix_api.requests import BitrixAPIRequest
from ..utils.functional import type_checker
from ..utils.types import Timeout
from ._base_scope import BaseScope

__all__ = [
    "Access",
]


class Access(BaseScope):
    """"""

    @type_checker
    def name(
            self,
            access: Iterable[Text],
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        if access.__class__ is not list:
            access = list(access)

        params = {
            "ACCESS": access,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.name,
            params=params,
            timeout=timeout,
        )



from dataclasses import dataclass
from typing import Text

from ..._constants import PYTHON_VERSION as _PV
from .base_bitrix_event import BaseBitrixEvent

_DATACLASS_KWARGS = {"eq": False, "frozen": True}

if _PV >= (3, 10):
    _DATACLASS_KWARGS["slots"] = True


@dataclass(**_DATACLASS_KWARGS)
class PortalDomainChangedEvent(BaseBitrixEvent):
    """"""
    old_domain: Text
    new_domain: Text



from b24pysdk.version import SDK_VERSION


def main():
    """Prints version information for python-telegram-bot, the Bot API and Python."""
    print(f"b24pysdk version {SDK_VERSION}")


if __name__ == "__main__":
    main()



from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Configurable",
]


class Configurable(BaseCRM):
    """The methods provide capabilities for working with configurable activities in timeline.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/configurable/index.html
    """

    @type_checker
    def add(
            self,
            *,
            owner_type_id: int,
            owner_id: int,
            fields: JSONDict,
            layout: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add configurable activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/configurable/crm-activity-configurable-add.html

        The method adds a configurable activity to the timeline.

        Args:
            owner_type_id: Identifier of the CRM entity type where the activity is created;

            owner_id: Identifier of the CRM element where the activity is created;

            fields: Object format:

                {
                    "typeId": 'value',

                    "completed": 'value',

                    "deadline": 'value',

                    "pingOffsets": 'value',

                    "isIncomingChannel": 'value',

                    "responsibleId": 'value',

                    "badgeCode": 'value',

                    "originatorId": 'value',

                    "originId": 'value',
                };

            layout: Associative array of a special structure describing the appearance of the activity in the timeline;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "ownerTypeId": owner_type_id,
            "ownerId": owner_id,
            "fields": fields,
            "layout": layout,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get configurable activity by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/configurable/crm-activity-configurable-get.html

        The method returns information about a configurable activity.

        Args:
            bitrix_id: Identifier od the activity;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update configurable activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/configurable/crm-activity-configurable-update.html

        The method makes changes to a configurable activity.

        Args:
            bitrix_id: Identifier of the activity;

            fields: Object format:

                {
                    "typeId": 'value',

                    "completed": 'value',

                    "deadline": 'value',

                    "pingOffsets": 'value',

                    "isIncomingChannel": 'value',

                    "responsibleId": 'value',

                    "badgeCode": 'value',

                    "originatorId": 'value',

                    "originId": 'value',
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id,
            fields,
            timeout=timeout,
        )



from functools import wraps
from typing import Callable, Dict, Final, Optional, Text

import requests

from ...error import (
    BitrixAPIInsufficientScope,
    BitrixAPIInvalidRequest,
    BitrixOAuthInsufficientScope,
    BitrixOAuthInvalidRequest,
    BitrixOAuthRequestError,
    BitrixOAuthRequestTimeout,
)
from ...utils.types import JSONDict, Number, Timeout
from ..protocols import BitrixOAuthProtocol
from ._base_requester import BaseRequester


def _bitrix_oauth_required(func: Callable):
    """"""
    @wraps(func)
    def wrapper(self: "BitrixOAuthRequester", *args, **kwargs):
        if not getattr(self, "_bitrix_oauth", None):
            raise AttributeError("'bitrix_oauth' is not defined")
        return func(self, *args, **kwargs)
    return wrapper


class BitrixOAuthRequester(BaseRequester):
    """"""

    _HEADERS: Final[Dict] = {"Content-Type": "application/x-www-form-urlencoded"}
    _OUATH_URL: Final[Text] = "https://oauth.bitrix.info/oauth/token/"
    _REST_URL: Final[Text] = "https://oauth.bitrix24.tech/rest/app.info/"

    __slots__ = ("_bitrix_oauth",)

    _bitrix_oauth: BitrixOAuthProtocol

    def __init__(
            self,
            bitrix_oauth: Optional[BitrixOAuthProtocol] = None,
            *,
            timeout: Timeout = None,
            max_retries: Optional[int] = None,
            initial_retry_delay: Optional[Number] = None,
            retry_delay_increment: Optional[Number] = None,
    ):
        super().__init__(
            timeout=timeout,
            max_retries=max_retries,
            initial_retry_delay=initial_retry_delay,
            retry_delay_increment=retry_delay_increment,
        )
        self._bitrix_oauth = bitrix_oauth

    @property
    def _headers(self) -> Dict:
        """"""
        return self._get_default_headers() | self._HEADERS

    def _request(self, url: Text, params: JSONDict) -> requests.Response:
        return requests.get(
            url=url,
            params=params,
            headers=self._headers,
            timeout=self._timeout,
        )

    def _get(self, url: Text, params: JSONDict) -> requests.Response:
        """"""

        try:
            return self._request_with_retries(url=url, params=params)

        except requests.Timeout as error:
            raise BitrixOAuthRequestTimeout(timeout=self._timeout, original_error=error) from error

        except requests.RequestException as error:
            raise BitrixOAuthRequestError(original_error=error) from error

    @classmethod
    def _parse_response(cls, response: requests.Response) -> JSONDict:
        """"""

        try:
            return super()._parse_response(response)

        except BitrixAPIInvalidRequest as error:
            raise BitrixOAuthInvalidRequest(response=error.response, json_response=error.json_response) from error

        except BitrixAPIInsufficientScope as error:
            raise BitrixOAuthInsufficientScope(response=error.response, json_response=error.json_response) from error

    @_bitrix_oauth_required
    def get_oauth_token(self, code: Text) -> JSONDict:
        """"""

        params = {
            "grant_type": "authorization_code",
            "client_id": self._bitrix_oauth.client_id,
            "client_secret": self._bitrix_oauth.client_secret,
            "code": code,
        }

        return self._parse_response(self._get(url=self._OUATH_URL, params=params))

    @_bitrix_oauth_required
    def refresh_oauth_token(self, refresh_token: Text) -> JSONDict:
        """"""

        params = {
            "grant_type": "refresh_token",
            "client_id": self._bitrix_oauth.client_id,
            "client_secret": self._bitrix_oauth.client_secret,
            "refresh_token": refresh_token,
        }

        return self._parse_response(self._get(url=self._OUATH_URL, params=params))

    def get_app_info(self, auth_token: Text) -> JSONDict:
        """"""

        params = {
            "auth": auth_token,
        }

        return self._parse_response(self._get(url=self._REST_URL, params=params))



from abc import ABC, abstractmethod
from functools import cached_property, wraps
from typing import TYPE_CHECKING, Callable, Dict, Final, Mapping, Optional, Sequence, Text, overload

from ..._config import Config
from ...error import BitrixAPIExpiredToken, BitrixResponse302JSONDecodeError
from ...utils.types import B24BatchMethods, B24BatchMethodTuple, JSONDict, Key, Timeout
from ..events import OAuthTokenRenewedEvent, PortalDomainChangedEvent
from ..functions import call_batch, call_batches, call_list, call_list_fast, call_method
from ..signals import BitrixSignalInstance
from .bitrix_app import AbstractBitrixApp, AbstractBitrixAppLocal
from .oauth_token import OAuthToken

if TYPE_CHECKING:
    from .oauth_placememt_data import OAuthPlacementData
    from .renewed_oauth_token import RenewedOAuthToken


def _bitrix_app_required(func: Callable):
    """"""
    @wraps(func)
    def wrapper(self: "AbstractBitrixToken", *args, **kwargs):
        if not getattr(self, "bitrix_app", None):
            raise AttributeError(f"'bitrix_app' is not defined for token of portal {self.domain}")
        return func(self, *args, **kwargs)
    return wrapper


class AbstractBitrixToken(ABC):
    """"""

    _AUTO_REFRESH: bool = True
    """"""

    domain: Text = NotImplemented
    """"""

    auth_token: Text = NotImplemented
    """"""

    refresh_token: Optional[Text] = NotImplemented
    """"""

    bitrix_app: Optional[AbstractBitrixApp] = NotImplemented
    """"""

    oauth_token_renewed_signal: BitrixSignalInstance = BitrixSignalInstance.create_signal(OAuthTokenRenewedEvent)
    """"""

    portal_domain_changed_signal: BitrixSignalInstance = BitrixSignalInstance.create_signal(PortalDomainChangedEvent)
    """"""

    @abstractmethod
    def __init__(self, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)

    def __str__(self):
        return f"<{'Webhook' if self.is_webhook else 'Application'} token of portal {self.domain}>"

    @classmethod
    def from_renewed_oauth_token(
            cls,
            renewed_oauth_token: "RenewedOAuthToken",
            bitrix_app: "AbstractBitrixApp",
    ) -> "AbstractBitrixToken":
        """"""
        oauth_token = renewed_oauth_token.oauth_token
        return cls(
            domain=renewed_oauth_token.portal_domain,
            auth_token=oauth_token.access_token,
            refresh_token=oauth_token.refresh_token,
            bitrix_app=bitrix_app,
        )

    @classmethod
    def from_oauth_placement_data(
            cls,
            oauth_placement_data: "OAuthPlacementData",
            bitrix_app: "AbstractBitrixApp",
    ) -> "AbstractBitrixToken":
        """"""
        oauth_token = oauth_placement_data.oauth_token
        return cls(
            domain=oauth_placement_data.domain,
            auth_token=oauth_token.access_token,
            refresh_token=oauth_token.refresh_token,
            bitrix_app=bitrix_app,
        )

    @property
    def is_webhook(self) -> bool:
        """"""
        return not bool(getattr(self, "bitrix_app", None))

    @property
    def _auth_data(self) -> Dict:
        """"""
        return dict(
            domain=self.domain,
            auth_token=self.auth_token,
            is_webhook=self.is_webhook,
            bitrix_token=self,
        )

    @cached_property
    def _config(self):
        return Config()

    @property
    def oauth_token(self) -> OAuthToken:
        """"""
        return OAuthToken(
            access_token=self.auth_token,
            refresh_token=self.refresh_token,
            expires=None,
        )

    @oauth_token.setter
    def oauth_token(self, oauth_token: OAuthToken):
        """"""
        self.auth_token = oauth_token.access_token
        self.refresh_token = oauth_token.refresh_token

    @_bitrix_app_required
    def get_oauth_token(self, code: Text) -> "RenewedOAuthToken":
        """"""
        return self.bitrix_app.get_oauth_token(code=code)

    @_bitrix_app_required
    def refresh_oauth_token(self) -> "RenewedOAuthToken":
        """"""
        return self.bitrix_app.refresh_oauth_token(refresh_token=self.refresh_token)

    def _refresh_and_set_oauth_token(self):
        """"""

        renewed_oauth_token = self.refresh_oauth_token()

        self.oauth_token = renewed_oauth_token.oauth_token

        self.oauth_token_renewed_signal.emit(OAuthTokenRenewedEvent(
            renewed_oauth_token=renewed_oauth_token,
        ))

    def _check_and_change_domain(self, new_domain: Text) -> bool:
        if not new_domain or new_domain == self.domain:
            return False

        old_domain = self.domain
        self.domain = new_domain

        self.portal_domain_changed_signal.emit(PortalDomainChangedEvent(
            old_domain=old_domain,
            new_domain=new_domain,
        ))

        return True

    def _call_with_retries(
            self,
            call_func: Callable,
            parameters: Dict,
    ) -> JSONDict:
        """"""

        try:
            return call_func(**self._auth_data, **parameters)

        except BitrixResponse302JSONDecodeError as error:
            if self._check_and_change_domain(error.new_domain):
                self._config.logger.info(
                    "Domain changed, retrying request",
                    context=dict(
                        old_domain=self.domain,
                        new_domain=error.new_domain,
                    ),
                )
                return call_func(**self._auth_data, **parameters)
            else:
                self._config.logger.warning(
                    "Caught BitrixResponse302JSONDecodeError, but domain did not change!",
                    context=dict(
                        old_domain=self.domain,
                        new_domain=error.new_domain,
                    ),
                )
                raise

        except BitrixAPIExpiredToken:
            if self._AUTO_REFRESH and not self.is_webhook:
                self._config.logger.info(
                    "Token expired, auto-refreshing token",
                    context=dict(
                        domain=self.domain,
                        auto_refresh=self._AUTO_REFRESH,
                        is_webhook=self.is_webhook,
                    ),
                )
                self._refresh_and_set_oauth_token()
                return call_func(**self._auth_data, **parameters)
            else:
                self._config.logger.warning(
                    "Token expired, but auto-refresh disabled!",
                    context=dict(
                        domain=self.domain,
                        is_webhook=self.is_webhook,
                        auto_refresh=self._AUTO_REFRESH,
                    ),
                )
                raise

    def call_method(
            self,
            api_method: Text,
            params: Optional[JSONDict] = None,
            *,
            timeout: Timeout = None,
            **kwargs,
    ) -> JSONDict:
        """"""
        return self._call_with_retries(
            call_func=call_method,
            parameters=dict(
                api_method=api_method,
                params=params,
                timeout=timeout,
                **kwargs,
            ),
        )

    @overload
    def call_batch(
            self,
            methods: Mapping[Key, B24BatchMethodTuple],
            *,
            halt: bool = False,
            ignore_size_limit: bool = False,
            timeout: Timeout = None,
            **kwargs,
    ) -> JSONDict: ...

    @overload
    def call_batch(
            self,
            methods: Sequence[B24BatchMethodTuple],
            *,
            halt: bool = False,
            ignore_size_limit: bool = False,
            timeout: Timeout = None,
            **kwargs,
    ) -> JSONDict: ...

    def call_batch(
            self,
            methods: B24BatchMethods,
            *,
            halt: bool = False,
            ignore_size_limit: bool = False,
            timeout: Timeout = None,
            **kwargs,
    ) -> JSONDict:
        """"""
        return self._call_with_retries(
            call_func=call_batch,
            parameters=dict(
                methods=methods,
                halt=halt,
                ignore_size_limit=ignore_size_limit,
                timeout=timeout,
                **kwargs,
            ),
        )

    @overload
    def call_batches(
            self,
            methods: Mapping[Key, B24BatchMethodTuple],
            *,
            halt: bool = False,
            timeout: Timeout = None,
            **kwargs,
    ) -> JSONDict: ...

    @overload
    def call_batches(
            self,
            methods: Sequence[B24BatchMethodTuple],
            *,
            halt: bool = False,
            timeout: Timeout = None,
            **kwargs,
    ) -> JSONDict: ...

    def call_batches(
            self,
            methods: B24BatchMethods,
            *,
            halt: bool = False,
            timeout: Timeout = None,
            **kwargs,
    ) -> JSONDict:
        """"""
        return self._call_with_retries(
            call_func=call_batches,
            parameters=dict(
                methods=methods,
                halt=halt,
                timeout=timeout,
                **kwargs,
            ),
        )

    def call_list(
            self,
            api_method: Text,
            params: Optional[JSONDict] = None,
            *,
            limit: Optional[int] = None,
            timeout: Timeout = None,
            **kwargs,
    ) -> JSONDict:
        """"""
        return self._call_with_retries(
            call_func=call_list,
            parameters=dict(
                api_method=api_method,
                params=params,
                limit=limit,
                timeout=timeout,
                **kwargs,
            ),
        )

    def call_list_fast(
            self,
            api_method: Text,
            params: Optional[JSONDict] = None,
            *,
            descending: bool = False,
            limit: Optional[int] = None,
            timeout: Timeout = None,
            **kwargs,
    ) -> JSONDict:
        """"""
        return self._call_with_retries(
            call_func=call_list_fast,
            parameters=dict(
                api_method=api_method,
                params=params,
                descending=descending,
                limit=limit,
                timeout=timeout,
                **kwargs,
            ),
        )


class AbstractBitrixTokenLocal(AbstractBitrixToken, ABC):
    """"""

    bitrix_app: AbstractBitrixAppLocal
    """"""

    @property
    def domain(self) -> Text:
        """"""
        return self.bitrix_app.domain

    @domain.setter
    def domain(self, domain: Text):
        """"""
        self.bitrix_app.domain = domain

    @classmethod
    def from_renewed_oauth_token(
            cls,
            renewed_oauth_token: "RenewedOAuthToken",
            bitrix_app: "AbstractBitrixAppLocal",
    ) -> "AbstractBitrixTokenLocal":
        """"""
        oauth_token = renewed_oauth_token.oauth_token
        return cls(
            auth_token=oauth_token.access_token,
            refresh_token=oauth_token.refresh_token,
            bitrix_app=bitrix_app,
        )

    @classmethod
    def from_oauth_placement_data(
            cls,
            oauth_placement_data: "OAuthPlacementData",
            bitrix_app: "AbstractBitrixAppLocal",
    ) -> "AbstractBitrixTokenLocal":
        """"""
        oauth_token = oauth_placement_data.oauth_token
        return cls(
            auth_token=oauth_token.access_token,
            refresh_token=oauth_token.refresh_token,
            bitrix_app=bitrix_app,
        )


class BitrixToken(AbstractBitrixToken):
    """"""

    def __init__(
            self,
            *,
            domain: Text,
            auth_token: Text,
            refresh_token: Optional[Text] = None,
            bitrix_app: Optional[AbstractBitrixApp] = None,
    ):
        self.domain = domain
        self.auth_token = auth_token
        self.refresh_token = refresh_token
        self.bitrix_app = bitrix_app


class BitrixTokenLocal(AbstractBitrixTokenLocal):
    """"""

    def __init__(
            self,
            *,
            auth_token: Text,
            refresh_token: Optional[Text] = None,
            bitrix_app: AbstractBitrixAppLocal,
    ):
        self.auth_token = auth_token
        self.refresh_token = refresh_token
        self.bitrix_app = bitrix_app


class BitrixWebhook(BitrixToken):
    """"""

    __AUTH_TOKEN_PARTS_COUNT: Final[int] = 2

    def __init__(
            self,
            *,
            domain: Text,
            auth_token: Text,
    ):
        super().__init__(domain=domain, auth_token=auth_token)

    @property
    def user_id(self) -> int:
        """"""

        auth_parts = self.auth_token.strip("/").split("/")

        if len(auth_parts) == self.__AUTH_TOKEN_PARTS_COUNT and auth_parts[0].isdigit():
            return int(auth_parts[0])
        else:
            raise ValueError(f"Invalid webhook auth_token format: expected 'user_id/webhook_key', got '{self.auth_token}'")

    @property
    def webhook_key(self) -> Text:
        """"""

        auth_parts = self.auth_token.strip("/").split("/")

        if len(auth_parts) == self.__AUTH_TOKEN_PARTS_COUNT:
            return auth_parts[1]
        else:
            raise ValueError(f"Invalid webhook auth_token format: expected 'user_id/webhook_key', got '{self.auth_token}'")



from abc import ABC, abstractmethod
from typing import Text

from ..requesters import BitrixOAuthRequester
from .renewed_oauth_token import RenewedOAuthToken


class AbstractBitrixApp(ABC):
    """"""

    client_id: Text = NotImplemented
    """"""

    client_secret: Text = NotImplemented
    """"""

    @abstractmethod
    def __init__(self, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)

    @property
    def is_local(self) -> bool:
        """"""
        return bool(getattr(self, "domain", None))

    def get_oauth_token(self, code: Text) -> RenewedOAuthToken:
        """"""
        return RenewedOAuthToken.from_dict(BitrixOAuthRequester(self).get_oauth_token(code))

    def refresh_oauth_token(self, refresh_token: Text) -> RenewedOAuthToken:
        """"""
        return RenewedOAuthToken.from_dict(BitrixOAuthRequester(self).refresh_oauth_token(refresh_token))


class AbstractBitrixAppLocal(AbstractBitrixApp, ABC):
    """"""

    domain: Text
    """"""


class BitrixApp(AbstractBitrixApp):
    """Local or market bitrix application"""

    def __init__(
            self,
            *,
            client_id: Text,
            client_secret: Text,
    ):
        self.client_id = client_id
        self.client_secret = client_secret


class BitrixAppLocal(AbstractBitrixAppLocal):
    """"""

    def __init__(
            self,
            *,
            domain: Text,
            client_id: Text,
            client_secret: Text,
    ):

        self.domain = domain
        self.client_id = client_id
        self.client_secret = client_secret



from .....bitrix_api.requests import BitrixAPIRequest
from .....utils.functional import type_checker
from .....utils.types import JSONDict, Timeout
from ..._base_crm import BaseCRM

__all__ = [
    "Blocks",
]


class Blocks(BaseCRM):
    """These methods offer capabilities for working with additional content blocks in an activity.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/layout-blocks/index.html
    """

    @type_checker
    def get(
            self,
            *,
            entity_type_id: int,
            entity_id: int,
            activity_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a set of additional content blocks in the activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/layout-blocks/crm-activity-layout-blocks-get.html

        This method allows a REST application to retrieve a set of additional content blocks in the activity that it has set.
        The REST application can only obtain the set of additional content blocks that it has established.

        Args:
            entity_type_id: Identifier of the CRM object type to which the activity is linked;

            entity_id: Identifier of the CRM object to which the activity is linked;

            activity_id: Identifier of the activity;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
            "activityId": activity_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def set(
            self,
            *,
            entity_type_id: int,
            entity_id: int,
            activity_id: int,
            layout: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Set a set of additional content blocks in the activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/layout-blocks/crm-activity-layout-blocks-set.html

        This method allows REST applications to set a set of additional content blocks in an activity.

        Args:
            entity_type_id: Identifier of the CRM object type to which the activity is linked;

            entity_id: Identifier of the CRM object to which the activity is linked;

            activity_id: Identifier of the activity;

            layout: Object describing the set of additional content blocks;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
            "activityId": activity_id,
            "layout": layout,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.set,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            *,
            entity_type_id: int,
            entity_id: int,
            activity_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete a set of additional content blocks in CRM activity.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/layout-blocks/crm-activity-layout-blocks-delete.html

        The method allows a REST application to delete a set of additional content blocks that it has installed for an activity.
        A REST application can only delete the set of additional content blocks that it has installed.

        Args:
            entity_type_id: Identifier of the CRM object type to which the activity is linked;

            entity_id: Identifier of the CRM object to which the activity is linked;

            activity_id: Identifier of the activity;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "entityTypeId": entity_type_id,
            "entityId": entity_id,
            "activityId": activity_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )



from functools import cached_property
from typing import Optional

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._base_crm import BaseCRM
from .entity import Entity

__all__ = [
    "Status",
]


class Status(BaseCRM):
    """The methods provide capabilities for managing elements from the reference book.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/status/index.html
    """

    @cached_property
    def entity(self) -> Entity:
        """"""
        return Entity(self)

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get description of CRM status fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/status/crm-status-fields.html

        The method returns a description of the fields in the directory.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new CRM directory element.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/status/crm-status-add.html

        The method creates a new element in the specified directory.

        Args:
            fields: Object format:

                {
                    "ENTITY_ID": "DEAL_STAGE",

                    "STATUS_ID": "DECISION",

                    "NAME": "Decision-Making",

                    "SORT": 70
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(fields=fields, timeout=timeout)

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the directory item by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/status/crm-status-get.html

        The method returns the directory item by ID.

        Args:
            bitrix_id:  Identifier of the directory item;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id=bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of directory items by filter.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/status/crm-status-list.html

        The  method returns a list of directory items on the filter.

        Args:
            filter: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = dict()

        if filter is not None:
            params["filter"] = filter

        if order is not None:
            params["order"] = order

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update an existing CRM directory item.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/status/crm-status-update.html

        This method updates an existing directory item.

        Args:
            bitrix_id: Identifier of the directory item;

            fields: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id=bitrix_id,
            fields=fields,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            params: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete CRM status element.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/status/crm-status-delete.html

        This method deletes a directory element.

        Args:
            bitrix_id: Identifier of the directory element;

            params: Set of parameters, where FORCED is a flag for forcibly deleting system elements;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        _params = {
            "id": bitrix_id,
        }

        if params is not None:
            _params["params"] = params

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=_params,
            timeout=timeout,
        )



from functools import cached_property
from typing import Iterable, Optional, Text

from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import JSONDict, Timeout
from .._productrows import Productrows
from .._relationships import Contact
from .._userfield import Userfield
from ..details import Details
from ..item.base_item import BaseItem
from .recurring import Recurring

__all__ = [
    "Deal",
]


class Deal(BaseItem):
    """The methods provide capabilities for managing deals.
    They allow you to retrieve fields, add, update, delete, and get lists of deals.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/index.html
    """

    @cached_property
    def contact(self) -> Contact:
        """"""
        return Contact(self)

    @cached_property
    def details(self) -> Details:
        """"""
        return Details(self)

    @cached_property
    def productrows(self) -> Productrows:
        """"""
        return Productrows(self)

    @cached_property
    def recurring(self) -> Recurring:
        """"""
        return Recurring(self)

    @cached_property
    def userfield(self) -> Userfield:
        """"""
        return Userfield(self)

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get deal fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/crm-deal-fields.html

        The method the description of deal fields, including custom ones.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            params: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Create a new deal.

        The method crm.deal.add creates a new deal.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/crm-deal-add.html

        Args:
            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            params: Object containing an additional set of parameters where

                - REGISTER_SONET_EVENT - whether to register the change event in the live feed 'Y' or not 'N';

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._add(
            fields,
            extra_params=params,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get deal by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/crm-deal-get.html

        The method returns a deal by its identifier.

        Args:
            bitrix_id: Identifier of the deal;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._get(bitrix_id, timeout=timeout)

    @type_checker
    def list(
            self,
            *,
            select: Optional[Iterable[Text]] = None,
            filter: Optional[JSONDict] = None,
            order: Optional[JSONDict] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of deals.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/crm-deal-list.html

        The method returns a list of deals based on a filter.

        Args:
            select: List of fields that should be populated for deals in the selection;

            filter: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'ASC' (ascending sort) or 'DESC' (descending sort);

            start: This parameter is used to manage pagination;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._list(
            select=select,
            filter=filter,
            order=order,
            start=start,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            fields: JSONDict,
            *,
            params: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update deal.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/crm-deal-update.html

        The method updates an existing deal.

        Args:
            bitrix_id: Identifier of the deal;

            fields: Object format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            params: Set of additional parameters where

                - REGISTER_SONET_EVENT - whether to register the change event in the activity stream 'Y' or not 'N',

                - REGISTER_HISTORY_EVENT - whether to create a record in history 'Y' or not 'N';

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._update(
            bitrix_id,
            fields,
            extra_params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete deal.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/deals/crm-deal-delete.html

        The method removes a deal and all associated objects.

        Deleting a deal will result in the removal of all related objects, such as CRM activities, history, Timeline activities, and others.

        Objects are deleted if they are not linked to other entities or elements. If the objects are linked to other entities, only the link to the deleted deal will be removed.

        Args:
            bitrix_id: Identifier of the deal;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id, timeout=timeout)




from .bitrix_api_batch_response import B24APIBatchResult, BitrixAPIBatchResponse
from .bitrix_api_list_response import BitrixAPIListFastResponse, BitrixAPIListResponse
from .bitrix_api_response import BitrixAPIResponse
from .bitrix_api_time_response import BitrixAPITimeResponse

__all__ = [
    "B24APIBatchResult",
    "BitrixAPIBatchResponse",
    "BitrixAPIListFastResponse",
    "BitrixAPIListResponse",
    "BitrixAPIResponse",
    "BitrixAPITimeResponse",
]



import threading
import typing

from .constants import DEFAULT_INITIAL_RETRY_DELAY, DEFAULT_MAX_RETRIES, DEFAULT_RETRY_DELAY_INCREMENT, DEFAULT_TIMEOUT
from .log import AbstractLogger, NullLogger
from .utils.types import DefaultTimeout, Number, Timeout


class _LocalConfig:
    """"""

    __slots__ = (
        "default_initial_retry_delay",
        "default_max_retries",
        "default_retry_delay_increment",
        "default_timeout",
        "logger",
    )

    default_initial_retry_delay: Number
    default_max_retries: int
    default_retry_delay_increment: Number
    default_timeout: DefaultTimeout
    logger: AbstractLogger

    def __init__(self):
        self.default_initial_retry_delay: Number = DEFAULT_INITIAL_RETRY_DELAY
        self.default_max_retries: int = DEFAULT_MAX_RETRIES
        self.default_retry_delay_increment: Number = DEFAULT_RETRY_DELAY_INCREMENT
        self.default_timeout: DefaultTimeout = DEFAULT_TIMEOUT
        self.logger = NullLogger()


class Config:
    """Thread-local configuration for SDK behavior"""

    __slots__ = ("_config",)

    _config: _LocalConfig

    _local_thread: threading.local = threading.local()

    def __init__(self):
        local_thread = type(self)._local_thread

        if not hasattr(local_thread, "config"):
            local_thread.config = _LocalConfig()

        self._config = local_thread.config

    def configure(
            self,
            *,
            default_initial_retry_delay: typing.Optional[Number] = None,
            default_max_retries: typing.Optional[int] = None,
            default_retry_delay_increment: typing.Optional[Number] = None,
            default_timeout: Timeout = None,
            logger: typing.Optional[AbstractLogger] = None,
            log_level: typing.Optional[int] = None,
    ):
        """"""

        if default_initial_retry_delay is not None:
            self.default_initial_retry_delay = default_initial_retry_delay

        if default_max_retries is not None:
            self.default_max_retries = default_max_retries

        if default_retry_delay_increment is not None:
            self.default_retry_delay_increment = default_retry_delay_increment

        if default_timeout is not None:
            self.default_timeout = default_timeout

        if logger is not None:
            self.logger = logger

        if log_level is not None:
            self.log_level = log_level

    @property
    def default_initial_retry_delay(self) -> Number:
        """Initial delay between retries in seconds"""
        return self._config.default_initial_retry_delay

    @default_initial_retry_delay.setter
    def default_initial_retry_delay(self, value: Number):
        """"""

        if not (isinstance(value, (int, float)) and value > 0):
            raise ValueError("Initial_retry_delay must be a positive number")

        self._config.default_initial_retry_delay = value

    @property
    def default_max_retries(self) -> int:
        """Maximum number retries that will occur when server is not responding"""
        return self._config.default_max_retries

    @default_max_retries.setter
    def default_max_retries(self, value: int):
        """"""

        if not (isinstance(value, int) and value >= 1):
            raise ValueError("Max_retries must be a positive integer (>= 1)")

        self._config.default_max_retries = value

    @property
    def default_retry_delay_increment(self) -> Number:
        """Amount by which delay between retries will increment after each retry"""
        return self._config.default_retry_delay_increment

    @default_retry_delay_increment.setter
    def default_retry_delay_increment(self, value: Number):
        """"""

        if not (isinstance(value, (int, float)) and value >= 0):
            raise ValueError("Retry_delay_increment must be a positive number")

        self._config.default_retry_delay_increment = value

    @property
    def default_timeout(self) -> DefaultTimeout:
        """Default timeout for API calls"""
        return self._config.default_timeout

    @default_timeout.setter
    def default_timeout(self, value: DefaultTimeout):
        """"""

        if not (isinstance(value, (int, float)) and value > 0):
            raise ValueError("Default_timeout must be a positive number or a tuple of two positive numbers (connect_timeout, read_timeout)")

        self._config.default_timeout = value

    @property
    def logger(self) -> AbstractLogger:
        """"""
        return self._config.logger

    @logger.setter
    def logger(self, value: AbstractLogger):
        """"""

        if not isinstance(value, AbstractLogger):
            raise TypeError("Logger must be an instance of AbstractLogger")

        self._config.logger = value

    @property
    def log_level(self) -> int:
        """"""
        return self._config.logger.level

    @log_level.setter
    def log_level(self, value: int):
        """"""

        logger_class = self.logger.__class__

        if value not in logger_class.LOG_LEVELS.values():
            raise ValueError(
                f"Invalid log level: {value}. "
                f"It must be one of the levels defined in {logger_class.__name__}.LOG_LEVELS: "
                f"{', '.join(map(str, logger_class.LOG_LEVELS.values()))}",
            )

        self._config.logger.set_level(value)



from typing import TYPE_CHECKING, Type

from psygnal import Signal, SignalInstance

if TYPE_CHECKING:
    from ..events import BaseBitrixEvent


class BitrixSignalInstance(SignalInstance):
    """"""

    @classmethod
    def create_signal(cls, event_class: Type["BaseBitrixEvent"]) -> Signal:
        return Signal(event_class, signal_instance_class=cls)



from .call import call
from .call_batch import call_batch
from .call_batches import call_batches
from .call_list import call_list
from .call_list_fast import call_list_fast
from .call_method import call_method

__all__ = [
    "call",
    "call_batch",
    "call_batches",
    "call_list",
    "call_list_fast",
    "call_method",
]



from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Mode",
]


class Mode(BaseCRM):
    """"""

    @type_checker
    def get(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            timeout=timeout,
        )



from ......bitrix_api.requests import BitrixAPIRequest
from ......utils.functional import type_checker
from ......utils.types import JSONDict, Timeout
from ...._base_crm import BaseCRM

__all__ = [
    "Field",
]


class Field(BaseCRM):
    """The methods provide capabilities for working with custom fields in the requisite template.

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/presets/fields/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get description of custom fields for the CRM requisite template.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/presets/fields/crm-requisite-preset-field-fields.html

        This method returns a formal description of the fields that define the customizable field of the requisite template.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)

    @type_checker
    def add(
            self,
            *,
            preset: JSONDict,
            fields: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add a custom field to the CRM requisite template

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/presets/fields/crm-requisite-preset-field-add.html

        This method adds a custom field to the requisite template.
        Before adding a user-defined field to the template it must be created.

        Args:
            preset: An object containing the identifier of the template to which the custom field is being added;

            fields: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "preset": preset,
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            bitrix_id: int,
            *,
            preset: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get custom field of requisite template by ID.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/presets/fields/crm-requisite-preset-field-get.html

        This method returns the description of a custom field of the requisite template by its identifier.

        Args:
            bitrix_id: Identifier of the custom field;

            preset: An object containing the identifier of the template from which the information about the custom field is extracted;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "preset": preset,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            preset: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get a list of customizable fields.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/presets/fields/crm-requisite-preset-field-list.html

        The method returns a list of all customizable fields for a specific requisites templates.

        Args:
            preset: An object containing the identifier value of the template from which the list of customizable fields is expected;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "preset": preset,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            bitrix_id: int,
            *,
            preset: JSONDict,
            fields: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Update custom fields of a given CRM requisite template.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/presets/fields/crm-requisite-preset-field-update.html

        This method updates a custom field in the requisite template.

        Args:
            bitrix_id: Identifier of the custom field to be updated;

            preset: An object containing the identifier of the template to which the custom field is added;

            fields: Object format:

                {

                    "FIELD_NAME": "value",

                    "FIELD_TITLE": "value",

                    "SORT": value,

                    "IN_SHORT_LIST": "value",

                }

                where FIELD_NAME is required;

            timeout: Timeout in seconds.

        Returns:
             Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "preset": preset,
            "fields": fields,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.update,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            preset: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Delete custom field from CRM requisite template.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/presets/fields/crm-requisite-preset-field-delete.html

        This method removes a custom field from the requisite template.

        Args:
            bitrix_id: Identifier of the custom field to be deleted;

            preset: An object containing the identifier of the template from which the custom field is being removed;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "preset": preset,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.delete,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def availabletoadd(
            self,
            *,
            preset: JSONDict,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get fields available for addition to the CRM requisite template.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/requisites/presets/fields/crm-requisite-preset-field-available-to-add.html

        This method returns the fields that can be added to the specified requisite template.

        Args:
            preset: An object containing the identifier value of the template for which you want to get the list of available customizable fields;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "preset": preset,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.availabletoadd,
            params=params,
            timeout=timeout,
        )



from functools import cached_property
from typing import Iterable, Optional, Text

from ...bitrix_api.requests import BitrixAPIRequest
from ...utils.functional import type_checker
from ...utils.types import JSONDict, Timeout
from .._base_scope import BaseScope
from .userfield import Userfield

__all__ = [
    "User",
]


class User(BaseScope):
    """"""

    @cached_property
    def userfield(self) -> Userfield:
        """"""
        return Userfield(self)

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._make_bitrix_api_request(
            api_wrapper=self.fields,
            timeout=timeout,
        )

    @type_checker
    def add(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=fields,
            timeout=timeout,
        )

    @type_checker
    def get(
            self,
            *,
            sort: Optional[Text] = None,
            order: Optional[Text] = None,
            filter: Optional[JSONDict] = None,
            admin_mode: Optional[bool] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        params = dict()

        if sort is not None:
            params["sort"] = sort

        if order is not None:
            params["order"] = order

        if filter is not None:
            params["filter"] = filter

        if admin_mode is not None:
            params["ADMIN_MODE"] = int(admin_mode)

        if start is not None:
            params["start"] = start

        return self._make_bitrix_api_request(
            api_wrapper=self.get,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def update(
            self,
            fields: JSONDict,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._make_bitrix_api_request(
            api_wrapper=self.update,
            params=fields,
            timeout=timeout,
        )

    @type_checker
    def current(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._make_bitrix_api_request(
            api_wrapper=self.current,
            timeout=timeout,
        )

    @type_checker
    def search(
            self,
            *,
            filter: Optional[JSONDict] = None,
            sort: Optional[Text] = None,
            order: Optional[Text] = None,
            admin_mode: Optional[bool] = None,
            start: Optional[int] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        params = dict()

        if filter is not None:
            params["filter"] = filter

        if sort is not None:
            params["sort"] = sort

        if order is not None:
            params["order"] = order

        if admin_mode is not None:
            params["ADMIN_MODE"] = int(admin_mode)

        if start is not None:
            params["start"] = start

        return self._make_bitrix_api_request(
            api_wrapper=self.search,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def admin(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""
        return self._make_bitrix_api_request(
            api_wrapper=self.admin,
            timeout=timeout,
        )

    @type_checker
    def access(
            self,
            access: Iterable[Text],
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """"""

        if access.__class__ is not list:
            access = list(access)

        params = {
            "ACCESS": access,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.access,
            params=params,
            timeout=timeout,
        )



from ....bitrix_api.requests import BitrixAPIRequest
from ....utils.functional import type_checker
from ....utils.types import Timeout
from .._base_crm import BaseCRM

__all__ = [
    "Communication",
]


class Communication(BaseCRM):
    """Method for working with system activities in the timeline

    Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/activity-base/index.html
    """

    @type_checker
    def fields(
            self,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get description of communication.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/timeline/activities/activity-base/crm-activity-communication-fields.html

        The method returns the description of communication for an activity.

        Args:
            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._fields(timeout=timeout)



from typing import Optional

from .....bitrix_api.requests import BitrixAPIRequest
from .....utils.functional import type_checker
from .....utils.types import JSONDict, Timeout
from ..._base_crm import BaseCRM

__all__ = [
    "Delivery",
]


class Delivery(BaseCRM):
    """These methods provide capabilities for managing deliveries in payments.
    
    Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/delivery-in-payment/index.html
    """

    @type_checker
    def add(
            self,
            *,
            payment_id: int,
            delivery_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Add delivery item to payment.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/delivery-in-payment/crm-item-payment-delivery-add.html

        This method adds a delivery item to the payment.

        Args:
            payment_id: Payment identifier;

            delivery_id: Delivery identifier;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "paymentId": payment_id,
            "deliveryId": delivery_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.add,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def list(
            self,
            *,
            payment_id: int,
            filter: JSONDict,
            order: Optional[JSONDict] = None,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Get the list of delivery items.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/delivery-in-payment/crm-item-payment-delivery-list.html

        This method retrieves the list of delivery items for a specific payment.

        Args:
            payment_id: Payment identifier.

            filter: Object in the format:

                {
                    field_1: value_1,

                    field_2: value_2,

                    ...,

                    field_n: value_n,
                };

            order: Object format:

                {
                    field_1: value_1,

                    ...,
                }

                where

                - field_n is the name of the field by which the selection will be sorted

                - value_n is a string value equals to 'asc' (ascending sort) or 'desc' (descending sort);

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "paymentId": payment_id,
            "filter": filter,
        }

        if order is not None:
            params["order"] = order

        return self._make_bitrix_api_request(
            api_wrapper=self.list,
            params=params,
            timeout=timeout,
        )

    @type_checker
    def delete(
            self,
            bitrix_id: int,
            *,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Remove delivery item from payment.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/delivery-in-payment/crm-item-payment-delivery-delete.html

        This method removes a delivery item from the payment.

        Args:
            bitrix_id: Identifier of the delivery item in the payment;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """
        return self._delete(bitrix_id, timeout=timeout)

    @type_checker
    def set_delivery(
            self,
            bitrix_id: int,
            *,
            delivery_id: int,
            timeout: Timeout = None,
    ) -> BitrixAPIRequest:
        """Reassign delivery item to another document.

        Documentation: https://apidocs.bitrix24.com/api-reference/crm/universal/payment/delivery-in-payment/crm-item-payment-delivery-set-delivery.html

        This method reassigns a delivery item to another delivery document.

        Args:
            bitrix_id: Identifier of the delivery item in the payment;

            delivery_id: Identifier of the delivery;

            timeout: Timeout in seconds.

        Returns:
            Instance of BitrixAPIRequest
        """

        params = {
            "id": bitrix_id,
            "deliveryId": delivery_id,
        }

        return self._make_bitrix_api_request(
            api_wrapper=self.set_delivery,
            params=params,
            timeout=timeout,
        )
